#########################################

About ExaQUte XMC

#########################################

The ExaQUte XMC library offers the user the ability to carry out uncertainty
quantification simulations using a multitude of Monte Carlo algorithms. Users 
of the library will be able to select one of many pre-programmed algorithms 
such as -

1. Simple Monte Carlo
2. Adaptive Monte Carlo
3. Simple Multi-Level Monte Carlo
4. Adaptive Multi-Level Monte Carlo
5. Contiuation Multi-Level Monte Carlo
6. Multi-Index Monte Carlo
7. Multi-Fidelity Monte Carlo
8. ...

as well as pre-programmed interfaces with widely used numerical analysis packages.

The library is also programmed in a highly modular way, allowing users to construct 
their own X-Monte Carlo (XMC) algorithm from several building block functions. It also 
allows them to write custom-interfaces easily for their own numerical analysis 
packages such as custom-finite element analysis or computational fluid dynamics
libraries. 

#########################################

ExaQUte XMC Library Structure

#########################################

The ExaQUte XMC library is organized as follows - 

1. The global directory 'xmc' is a python package that contains all the
packages, subpackages and modules of the library.

2. Within this folder, each file contains the definition of one class.
Every file is treated as a module that can be imported. If the class is
called ClassName, the file is named className.py

3. Inside each class, there are multiple types of members. They are as
follows
	- Instances of classes
	- Instances of function definition objects.
	- Variables and other data structures
	- Method definitions

4. The instantiation of function definition objects is used to enable a
function to have multiple definitions. For example, we require that the
optimalHierarchy method of the HierarchyOptimiser class to be called as
optimalHierarchy(), but to evaluate different expressions for the
optimal number of samples and levels based on the type of MLMC
algorithm. 

In this way	the user can simply specify the type of algorithm they
would like to run and the definition of optimalHierarchy automatically
changes based on this, without any further intervention.

This is achieved by instantiating optimalHierarchy to be of the type
specificOptimalHierarchyDefinition through the constructor of the
HierarchyOptimiser class as follows - 

	class HierarchyOptimiser(): 
		def __init__(self): 
			...
			self.optimalHierarchy = specificOptimalHierarchyDefinition 
			...

In this way, every call to optimalHierarchy() in the code is
replaced by specificOptimalHierarchyDefinition. 

5. It is hence implied that every "general" method such as
optimalHierarchy will have a corresponding list of specific definitions,
one of which is selected at runtime during the construction of the class
containing the method. 

To organize this, every className.py file has a corresponding
methodDefs_className folder. Inside this folder, there are multiple
files of name generalMethod.py. Within each generalMethod.py, there are
a list of definitions as follows -

	# generalMethod.py

	def specificDefinition_1(): ...

	def specificDefinition_2(): ...

In className.py, these definitions are imported. Based on the
type of algorithm required, the constructor of the class then sets
self.generalMethod = specificDefinition_1/2/...

As an example, we consider again the optimalHierarchy method. Inside
HierarchyOptimiser.py, the definitions are imported as 

	import methodDefs_hierarchyOptimiser.optimalHierarchy as mdoh

	class HierarchyOptimiser():
		def __init__(self):
			self.optimalHierarchy = mdoh.specificDefinition1

7. Imports are carried out as follows. For function definition objects,
the following is used. 

	import methodDefs_className.generalMethod as mdgm

The alias used for the imported module always begins with md and is then
followed by the first letter of each word involved in the general
method's name. For class definition imports, a similar alias convention
is used. 

	import className as cn


#########################################

Instructions for tutorial files

#########################################

General Instructions:

All tutorial files are to be run as follows

    python3 test_*.py 

No additional arguments are required. If the user has the PyCOMPSs scheduling
tool installed and the imports in the required files appropriately set, then 
the test files can be run as follows

    runcompss --options=value /global/path/to/test_*.py

1. example_mc.py

This test is for a simple Monte Carlo algorithm that estimates the mean
of a standard normal variable. The algorithm begins with a prescribed number
of samples. It estimates the mean and the error in the mean estimate. If the
error does not satisfy a prescribed tolerance, it doubles the number of samples
and repeats the process until convergence.

2. example_amc.py    

This test is for an adaptive Monte Carlo algorithm that estimates the mean
of a standard normal variable. The algorithm begins with a prescribed number
of samples. It estimates the mean and the error in the mean estimate. If the 
error does not satisfy a final tolerance, it then computes a new number of 
samples that will be required to satisfy the tolerance requirement based on
the assumption of asymptotic normality, and repeats this procedure until 
convergence.

3. example_mlmc.py

This test is for a simple Multi-Level Monte Carlo algorithm that estimates the mean
of a standard normal variable. A hierarchy of normal random generators whose, the 
mean and variance of the differences of whose samples decay geometrically with increasing 
level towards 0, are used to generate samples. The algorithm begins with a prescribed number
of samples and levels. It estimates the mean and the error in the mean estimate. If the
error does not satisfy a prescribed tolerance, it increases the number of levels by one,
doubles the number of samples in all existing levels, and adds a default number of samples 
to the new level. This process is repeated until convergence.

4. example_amlmc.py

This test is for an adaptive Multi-Level Monte Carlo algorithm that estimates the mean
of a standard normal variable. A hierarchy of normal random generators whose, the 
mean and variance of the differences of whose samples decay geometrically with increasing 
level towards 0, are used to generate samples. The algorithm begins with a prescribed number
of samples and levels. It estimates the mean and the error in the mean estimate. It then fits
geometric models for the bias, variance and cost across levels on the computed data. If the
error does not satisfy a prescribed tolerance, it increases the number of levels by one and 
computes the cost-optimal number of samples for each level using level-wise bias, variance and 
cost estimates as well as the models for these three quantities. to the new level. This 
process is repeated until convergence.


#########################################

Dependencies

#########################################

Depends on NumPy

Scheduler tool: PyCOMPSs URL:https://compss.bsc.es/

Recommended solver, used in provided scripts: Kratos DOI:10.5281/zenodo.3234645
