# Import Kratos
import KratosMultiphysics

# Import Kratos problem
from xmc.classDefs_solverWrapper.problemDefs_KratosMultiphysics.poisson_2d.simulation_definition import SimulationScenario
import xmc.classDefs_solverWrapper.methodDefs_KratosSolverWrapper.solve as mds

# Import cpickle to pickle the serializer
try:
    import cpickle as pickle  # Use cPickle on Python 2.7
except ImportError:
    import pickle

# Import xmc classes
import xmc.solverWrapper as sw


class KratosSolverWrapper(sw.SolverWrapper):

    def __init__(self,**keywordArgs):
        super().__init__(**keywordArgs)
        self.analysis = SimulationScenario
        self.number_qoi = 1

        self.refinement_strategy = "concurrent_adaptive_refinement"
        # self.refinement_strategy = "unique_adaptive_refinement"
        # self.refinement_strategy = "reading_from_file"

        if (self.solverWrapperIndex == []):
            # reading of json files
            self.project_parameters_path = "xmc/classDefs_solverWrapper/problemDefs_KratosMultiphysics/poisson_2d/problem_settings/parameters_poisson_square_2d_finer.json"
            self.refinement_parameters_path = "xmc/classDefs_solverWrapper/problemDefs_KratosMultiphysics/poisson_2d/problem_settings/parameters_refinement.json"
            # serialization
            self.is_project_parameters_pickled = False
            self.is_model_pickled = False
            self.SerializeModelParameters()
            self.SetRefinementParameters()
            self.SerializeRefinementParameters()
            self.ComputeMeshParameters()

        # Multi Level Monte Carlo algorithm
        else:
            if (self.refinement_strategy == "concurrent_adaptive_refinement"):
                # reading of json files
                self.project_parameters_path = "xmc/classDefs_solverWrapper/problemDefs_KratosMultiphysics/poisson_2d/problem_settings/parameters_poisson_square_2d_finer.json"
                self.refinement_parameters_path = "xmc/classDefs_solverWrapper/problemDefs_KratosMultiphysics/poisson_2d/problem_settings/parameters_refinement.json"
                # serialization
                self.is_project_parameters_pickled = False
                self.is_model_pickled = False
                self.is_custom_settings_metric_refinement_pickled = False
                self.is_custom_settings_remesh_refinement_pickled = False
                self.SerializeModelParameters()
                self.SetRefinementParameters()
                self.SerializeRefinementParameters()
                self.ComputeMeshParameters()

            else:
                raise Exception ("Select KratosMultiphysics refinement stategy.\nOptions:\
                    \n i) concurrent_adaptive_refinement")

    def solve(self,random_variable):
        time_for_qoi = 0.0
        if all([component>=0 for component in self.solverWrapperIndex]):
            # local variables
            pickled_model = self.pickled_model
            pickled_project_parameters = self.pickled_project_parameters
            current_analysis = self.analysis
            if (self.number_qoi == 1):
                qoi,time_for_qoi = self.ExecuteInstance(random_variable)
                qoi_list = [qoi] # list of future.compss objects
        else:
            qoi,time_for_qoi = mds.returnZeroQoIandTime_Task()
            qoi_list = [qoi]*self.number_qoi
        return qoi_list, time_for_qoi

    """
    function executing an instance of the UQ algorithm, i.e. a single MC simulation and eventually the refinement (that occurs before the simulation run)
    input:  self: an instance of the class
    output: qoi: quantity of interest value
    """
    def ExecuteInstance(self,random_variable):
        current_index = self.solverWrapperIndex[0]
        # local variables
        pickled_coarse_model = self.pickled_model
        pickled_coarse_project_parameters = self.pickled_project_parameters
        mesh_sizes = self.mesh_sizes
        recursive_maximal_size = self.recursive_maximal_size # check if setting or not maximal mesh size
        pickled_custom_metric_refinement_parameters = self.pickled_custom_metric_refinement_parameters
        pickled_custom_remesh_refinement_parameters = self.pickled_custom_remesh_refinement_parameters
        current_analysis = self.analysis
        time_for_qoi = 0.0
        if (current_index == 0):
            qoi,pickled_current_model,pickled_current_project_parameters,time_for_qoi = \
                mds.ExecuteInstanceAux_Task(current_index,pickled_coarse_model,pickled_coarse_project_parameters,pickled_custom_metric_refinement_parameters,pickled_custom_remesh_refinement_parameters,mesh_sizes,recursive_maximal_size,random_variable,current_analysis,time_for_qoi)
        else:
            for index_iterator in range(current_index+1):
                qoi,pickled_current_model,pickled_current_project_parameters,time_for_qoi = \
                    mds.ExecuteInstanceAux_Task(index_iterator,pickled_coarse_model,pickled_coarse_project_parameters,pickled_custom_metric_refinement_parameters,pickled_custom_remesh_refinement_parameters,mesh_sizes,recursive_maximal_size,random_variable,current_analysis,time_for_qoi)
                del(pickled_coarse_model,pickled_coarse_project_parameters)
                pickled_coarse_model = pickled_current_model
                pickled_coarse_project_parameters = pickled_current_project_parameters
                del(pickled_current_model,pickled_current_project_parameters)
        return qoi,time_for_qoi

    """
    function serializing and pickling the model and the project parameters of the problem
    the serialization-pickling process is the following:
    i)   from Model/Parameters Kratos object to StreamSerializer Kratos object
    ii)  from StreamSerializer Kratos object to pickle string
    iii) from pickle string to StreamSerializer Kratos object
    iv)  from StreamSerializer Kratos object to Model/Parameters Kratos object
    requires: self.project_parameters_path: path of the Project Parameters file
    builds: self.pickled_model:              pickled model
            self.pickled_project_parameters: pickled project parameters
    input:  self: an instance of the class
    """
    def SerializeModelParameters(self):
        with open(self.project_parameters_path,'r') as parameter_file:
            parameters = KratosMultiphysics.Parameters(parameter_file.read())
        model = KratosMultiphysics.Model()
        fake_sample = [1.0]
        simulation = self.analysis(model,parameters,fake_sample)
        simulation.Initialize()
        serialized_model = KratosMultiphysics.StreamSerializer()
        serialized_model.Save("ModelSerialization",simulation.model)
        serialized_project_parameters = KratosMultiphysics.StreamSerializer()
        serialized_project_parameters.Save("ParametersSerialization",simulation.project_parameters)
        self.serialized_model = serialized_model
        self.serialized_project_parameters = serialized_project_parameters
        # pickle dataserialized_data
        pickled_model = pickle.dumps(serialized_model, 2) # second argument is the protocol and is NECESSARY (according to pybind11 docs)
        pickled_project_parameters = pickle.dumps(serialized_project_parameters, 2)
        self.pickled_model = pickled_model
        self.pickled_project_parameters = pickled_project_parameters
        self.is_project_parameters_pickled = True
        self.is_model_pickled = True
        print("\n","#"*50," SERIALIZATION MODEL PART AND PARAMETERS COMPLETED ","#"*50,"\n")

    """
    function reading the refinement parameters passed from json file
    input:  self: an instance of the class
    """
    def SetRefinementParameters(self):
        with open(self.refinement_parameters_path,'r') as parameter_file:
            parameters = KratosMultiphysics.Parameters(parameter_file.read())
        self.custom_metric_refinement_parameters = parameters["hessian_metric"]
        self.custom_remesh_refinement_parameters = parameters["refinement_mmg"]
        self.custom_simulation_refinement_parameters = parameters["simulation_refinement"]
        self.recursive_maximal_size = self.custom_simulation_refinement_parameters["recursive_maximal_size"].GetString()

    """
    function serializing and pickling the custom setting metric and remeshing for the refinement
    requires: self.custom_metric_refinement_parameters
              self.custom_remesh_refinement_parameters
    builds: self.pickled_custom_metric_refinement_parameters
            self.pickled_custom_remesh_refinement_parameters
    input:  self: an instance of the class
    """
    def SerializeRefinementParameters(self):
        metric_refinement_parameters = self.custom_metric_refinement_parameters
        remeshing_refinement_parameters = self.custom_remesh_refinement_parameters
        # save parameters as StreamSerializer Kratos objects
        serialized_metric_refinement_parameters = KratosMultiphysics.StreamSerializer()
        serialized_metric_refinement_parameters.Save("MetricRefinementParametersSerialization",metric_refinement_parameters)
        serialized_remesh_refinement_parameters = KratosMultiphysics.StreamSerializer()
        serialized_remesh_refinement_parameters.Save("RemeshRefinementParametersSerialization",remeshing_refinement_parameters)
        # pickle parameters
        pickled_metric_refinement_parameters = pickle.dumps(serialized_metric_refinement_parameters, 2) # second argument is the protocol and is NECESSARY (according to pybind11 docs)
        pickled_remesh_refinement_parameters = pickle.dumps(serialized_remesh_refinement_parameters, 2) # second argument is the protocol and is NECESSARY (according to pybind11 docs)
        self.pickled_custom_metric_refinement_parameters = pickled_metric_refinement_parameters
        self.pickled_custom_remesh_refinement_parameters = pickled_remesh_refinement_parameters
        print("\n","#"*50," SERIALIZATION REFINEMENT PARAMETERS COMPLETED ","#"*50,"\n")

    """
    function giving as output the mesh discretization parameter
    the mesh parameter is the reciprocal of the minimum mesh size of the grid
    h_lev=h_0*M^(-lev)
    input:  self: an instance of the class
    """
    def ComputeMeshParameters(self):
        self.mesh_sizes = []
        self.mesh_parameters = []
        h0 = self.custom_simulation_refinement_parameters["initial_mesh_size"].GetDouble()
        M  = self.custom_simulation_refinement_parameters["mesh_refinement_coefficient"].GetDouble()
        if (self.solverWrapperIndex == []):
            self.solverWrapperIndex.append(0)
        for index in range(0,self.solverWrapperIndex[0]+1):
            h_current_index = h0 * M**(-index) # h_{lev=l} = h_{lev=0}*coefficient**-(index)
            mesh_parameter_current_index = h_current_index**(-1)
            self.mesh_sizes.append(h_current_index)
            self.mesh_parameters.append(mesh_parameter_current_index)
