# Import Kratos
import KratosMultiphysics
from KratosMultiphysics.MultilevelMonteCarloApplication.adaptive_refinement_utilities import AdaptiveRefinement

# Import PyCOMPSs
from exaqute.ExaquteTaskPyCOMPSs import *   # to execute with runcompss
# from exaqute.ExaquteTaskHyperLoom import *  # to execute with the IT4 scheduler
# from exaqute.ExaquteTaskLocal import *      # to execute with python3

# Import python packages
import time 

# Import cpickle to pickle the serializer
try:
    import cpickle as pickle  # Use cPickle on Python 2.7
except ImportError:
    import pickle


"""
function evaluating the QoI and the cost of the simulation, computing the mesh of current index self.solverWrapperIndex[0]
refining recursively from the coarsest mesh
input:
        current_index:                     current Multilevel Monte Carlo index we are solving
        pickled_coarse_model:              pickled model
        pickled_coarse_project_parameters: pickled parameters
        mesh_sizes:                        mesh sizes for indexes of interest
output:
        qoi:                              quantity of interest
        pickled_finer_model:              pickled finer model
        pickled_finer_project_parameters: pickled finer parameters
"""
@ExaquteTask(returns=4)
def ExecuteInstanceAux_Task(current_index,pickled_coarse_model,pickled_coarse_project_parameters,pickled_custom_metric_refinement_parameters,pickled_custom_remesh_refinement_parameters,mesh_sizes,recursive_maximal_size,random_variable,current_analysis_stage,previous_computational_time):
    start_time = time.time()
    # unpickle model and build Kratos Model object
    serialized_model = pickle.loads(pickled_coarse_model)
    current_model = KratosMultiphysics.Model()
    serialized_model.Load("ModelSerialization",current_model)
    del(serialized_model)
    # unpickle parameters and build Kratos Parameters object
    serialized_project_parameters = pickle.loads(pickled_coarse_project_parameters)
    current_project_parameters = KratosMultiphysics.Parameters()
    serialized_project_parameters.Load("ParametersSerialization",current_project_parameters)
    del(serialized_project_parameters)
    # refine if current current_index > 0, adaptive refinement based on the solution of previous index
    if (current_index > 0):
        # unpickle metric and remesh refinement parameters and build Kratos Parameters objects
        serialized_custom_metric_refinement_parameters = pickle.loads(pickled_custom_metric_refinement_parameters)
        serialized_custom_remesh_refinement_parameters = pickle.loads(pickled_custom_remesh_refinement_parameters)
        current_custom_metric_refinement_parameters = KratosMultiphysics.Parameters()
        current_custom_remesh_refinement_parameters = KratosMultiphysics.Parameters()
        serialized_custom_metric_refinement_parameters.Load("MetricRefinementParametersSerialization",current_custom_metric_refinement_parameters)
        serialized_custom_remesh_refinement_parameters.Load("RemeshRefinementParametersSerialization",current_custom_remesh_refinement_parameters)
        del(serialized_custom_metric_refinement_parameters,serialized_custom_remesh_refinement_parameters)
        # set minimal and maximal mesh sizes
        minimal_mesh_size_index = mesh_sizes[current_index]
        if (recursive_maximal_size == "True"):
            maximal_mesh_size_index = mesh_sizes[current_index-1]
        else:
            maximal_mesh_size_index = 10.0
        # refine the model Kratos object
        adaptive_refinement_manager = AdaptiveRefinement(current_model,current_project_parameters,current_custom_metric_refinement_parameters,current_custom_remesh_refinement_parameters,minimal_mesh_size_index,maximal_mesh_size_index)
        refined_model,refined_project_parameters = adaptive_refinement_manager.ComputeAdaptiveRefinement()
        # initialize the model Kratos object
        simulation = current_analysis_stage(refined_model,refined_project_parameters,random_variable)
        simulation.Initialize()
        # update model Kratos object
        current_model = simulation.model
        current_project_parameters = simulation.project_parameters
        del(simulation)
    simulation = current_analysis_stage(current_model,current_project_parameters,random_variable)
    simulation.Run()
    qoi = simulation.EvaluateQuantityOfInterest()
    # save model and parameters as StreamSerializer Kratos objects
    serialized_finer_model = KratosMultiphysics.StreamSerializer()
    serialized_finer_model.Save("ModelSerialization",simulation.model)
    serialized_finer_project_parameters = KratosMultiphysics.StreamSerializer()
    serialized_finer_project_parameters.Save("ParametersSerialization",simulation.project_parameters)
    # pickle model and parameters
    pickled_finer_model = pickle.dumps(serialized_finer_model, 2) # second argument is the protocol and is NECESSARY (according to pybind11 docs)
    pickled_finer_project_parameters = pickle.dumps(serialized_finer_project_parameters, 2) # second argument is the protocol and is NECESSARY (according to pybind11 docs)
    del(simulation)
    end_time = time.time()
    computational_time = previous_computational_time + (end_time-start_time)
    return qoi,pickled_finer_model,pickled_finer_project_parameters,computational_time

@ExaquteTask(returns=2)
def returnZeroQoIandTime_Task():
    qoi = 0.0
    time_for_qoi = 0.0
    return qoi,time_for_qoi
