# Import PyCOMPSs (if installed)
# from exaqute.ExaquteTaskPyCOMPSs import *   # to execute with runcompss

# Import empty definitions (if PyCOMPSs is not installed)
from scheduler.ExaquteTaskLocal import *      # to execute with python3

# Import XMC methods
from xmc.tools import dynamicImport
from xmc.tools import mergeTwoListsIntoOne

class HierarchyOptimiser():
    """
    This class is in charge of deciding the best hierarchy. It is able to:
    - provide the best index set (according to its implement meaning of
            best) within adjustable constraints;
    - give the tolerance splitting associated to that choice (be it fixed or
            computed)
    - provide the sample numbers per Monte Carlo index that minimise the
            cost.

    It is assumed here that the hierarchy is optimised with respect to a single QoI.
    """

    def __init__(self,**keywordArgs):
        # Attributes
        self.indexSpace = keywordArgs.get("indexSpace",None)
        self.toleranceSplittingBounds = keywordArgs.get("toleranceSplittingBounds",None)
        self.defaultHierarchy = keywordArgs.get('defaultHierarchy',None)

        # Methods
        self.optimalIndexSet = dynamicImport(keywordArgs.get("optimalIndexSet"))
        self.optimalSampleNumbers = dynamicImport(keywordArgs.get("optimalSampleNumbers"))
        self.updateHierarchySpace = dynamicImport(keywordArgs.get("updateHierarchySpace", "xmc.tools.doNothing"))
        self.toleranceSplittingMethod = dynamicImport(keywordArgs.get("toleranceSplittingMethod", "xmc.tools.doNothing"))

    def inputDictionaryTemplate(self):
        """
        Supply an empty default dictionary whose entries are required by the sample number
        and index set calculation methods.
        """
        input_dict = {
            'oldHierarchy':None,
            'newSampleNumber':None,
            'cdfValue':None,
            'splittingParameter':None,
            'tolerances':None,
            'models':None,
            'parametersForModel':None,
            'estimations':None,
            'costModel':None,
            'costParameters':None,
            'costEstimations':None
        }
        return input_dict

    def toleranceSplitting(self, inputDict):
        """
        Compute the tolerance splitting parameter as follows - 
        If the upper and lower bounds for the splitting parameter are the same, return this value as the splitting parameter.
        If not, then apply these upper and lower bounds to the splitting parameter contained in the input dictionary
        """
        assert(len(inputDict['tolerances'])<=1),"toleranceSplitting() cannot handle more than one tolerance for now"

        if self.toleranceSplittingBounds[0]==self.toleranceSplittingBounds[1]:
            inputDict['splittingParameter'] = self.toleranceSplittingBounds[0]
        else:
            splitting_parameter = self.toleranceSplittingMethod(inputDict)
            splitting_parameter = min(max(splitting_parameter,self.toleranceSplittingBounds[0]),self.toleranceSplittingBounds[1])
            inputDict['splittingParameter'] = splitting_parameter
        return inputDict

    def optimalHierarchy(self, inputDict):
        """
        Compute the best hierarchy by individually computing the best index
        set and best number of samples
        """
        inputDict = self.toleranceSplitting(inputDict)
        indices = self.optimalIndexSet(inputDict)
        sample_numbers = self.optimalSampleNumbers(inputDict, indices)

        old_hierarchy = inputDict['oldHierarchy']
        new_hierarchy = []
        if (len(old_hierarchy[0]) == 0):
            new_hierarchy = [indices,sample_numbers]
        else:
            new_hierarchy = mergeTwoListsIntoOne(indices,sample_numbers)

        return new_hierarchy
