import numpy as np
import sys

def singleIndexDoubleSampleNumber(inputDict, newLevels):
    """
    Double the existing number of samples
    """
    oldHierarchy = inputDict['oldHierarchy']
    old_sample_number = oldHierarchy[0][1]
    return [2*old_sample_number]

def singleIndexAdaptive(inputDict, newLevels):
    """
    Compute optimal number of samples needed to achieve a prescribed 
    tolerance based on the Central Limit Theorem and asymptotic normality
    """
    oldHierarchy = inputDict['oldHierarchy']
    variance = inputDict['estimations'][0][0]*oldHierarchy[0][1]
    cdf_value = inputDict['cdfValue']
    tolerance = inputDict['tolerances'][0]
    return [int(np.ceil(variance*(cdf_value/tolerance)**2))]

def multiLevelDoubleAllSamples(inputDict, newLevels):
    """
    Returns a list of sample numbers of same length as the number of entries
    in newLevels. Doubles the number of samples from oldHierarchy if an
    entry of newLevels exists in oldHierarchy. If not, allocate a default 
    newSampleNumber to the entry
    """
    oldHierarchy = inputDict['oldHierarchy']
    newSampleNumber = inputDict['newSampleNumber']
    new_samples = []
    for newLevel in newLevels:
        is_level_found = False
        for oldElement in oldHierarchy:
            if newLevel==oldElement[0]:
                new_samples.append(2*oldElement[1])
                is_level_found = True
                break
        if is_level_found is False:        
            new_samples.append(newSampleNumber)
    return new_samples

def multiLevelCostMinimisationTErr(inputDict, newLevels):
    """
    Compute the cost optimal distribution of samples over levels that minimizes
    the total CPU time for total error to satisfy a given tolerance.  
    """
    oldHierarchy = inputDict['oldHierarchy']
    tolerance = inputDict['tolerances'][0]
    cost_model = inputDict['costModel']
    cost_parameters = inputDict['costParameters']
    cost_estimations = inputDict['costEstimations'] 
    variance_model = inputDict['models'][1]
    variance_parameters = inputDict['parametersForModel'][1]
    variance_estimations = [inputDict['estimations'][1][i]*oldHierarchy[i][1] for i in range(len(oldHierarchy))]

    new_samples = []
    max_level = max([newLevels[i][0] for i in range(len(newLevels))])

    # Compute this term - sum_{k=0}^L sqrt(C_k V_k) for later use
    cost_variance_sum = 0.0
    for i in range(len(newLevels)):
        is_level_found = False
        for k in range(len(oldHierarchy)):
            if(oldHierarchy[k][0][0]==newLevels[i][0]):
                cost_level = cost_estimations[k]
                variance_level = variance_estimations[k]
                is_level_found = True
                break
        if is_level_found is False:
            cost_level = cost_model(cost_parameters,newLevels[i])
            variance_level = variance_model(variance_parameters,newLevels[i])
        cost_variance_sum += np.sqrt(cost_level*variance_level)

    # Compute number of samples
    for i in range(len(newLevels)):
        is_level_found = False
        for k in range(len(oldHierarchy)):
            if(oldHierarchy[k][0][0]==newLevels[i][0]):
                cost_level = cost_estimations[k]
                variance_level = variance_estimations[k]
                is_level_found = True
                break
        if is_level_found is False:
            cost_level = cost_model(cost_parameters,newLevels[i])
            variance_level = variance_model(variance_parameters,newLevels[i])
        new_sample = (cdf_value/(splitting_parameter*tolerance))**2 * cost_variance_sum * np.sqrt(variance_level/cost_level)
        new_samples.append(int(np.ceil(new_sample)))
    return new_samples

def multiLevelCostMinimisationMSE(inputDict,newLevels):
    """
    Compute the cost optimal distribution of samples over levels that minimizes
    the total CPU time for mean-squared error to satisfy a given tolerance 
    """
    oldHierarchy = inputDict['oldHierarchy']
    cdf_value = inputDict['cdfValue']
    splitting_parameter = inputDict['splittingParameter']
    tolerance = inputDict['tolerances'][0]
    cost_model = inputDict['costModel']
    cost_parameters = inputDict['costParameters']
    cost_estimations = inputDict['costEstimations'] 
    variance_model = inputDict['models'][1]
    variance_parameters = inputDict['parametersForModel'][1]
    variance_estimations = [inputDict['estimations'][1][i]*oldHierarchy[i][1] for i in range(len(oldHierarchy))]

    new_samples = []
    max_level = max([newLevels[i][0] for i in range(len(newLevels))])

    # Compute this term - sum_{k=0}^L sqrt(C_k V_k) for later use
    cost_variance_sum = 0.0
    for i in range(len(newLevels)):
        is_level_found = False
        for k in range(len(oldHierarchy)):
            if(oldHierarchy[k][0][0]==newLevels[i][0]):
                cost_level = cost_estimations[k]
                variance_level = variance_estimations[k]
                is_level_found = True
                break
        if is_level_found is False:
            cost_level = cost_model(cost_parameters,newLevels[i])
            variance_level = variance_model(variance_parameters,newLevels[i])
        cost_variance_sum += np.sqrt(cost_level*variance_level)

    # Compute number of samples
    for i in range(len(newLevels)):
        is_level_found = False
        for k in range(len(oldHierarchy)):
            if(oldHierarchy[k][0][0]==newLevels[i][0]):
                cost_level = cost_estimations[k]
                variance_level = variance_estimations[k]
                is_level_found = True
                break
        if is_level_found is False:
            cost_level = cost_model(cost_parameters,newLevels[i])
            variance_level = variance_model(variance_parameters,newLevels[i])
        new_sample = (cdf_value/(splitting_parameter*tolerance))**2 * cost_variance_sum * np.sqrt(variance_level/cost_level)
        new_samples.append(int(np.ceil(new_sample)))
    return new_samples
