# XMC imports
import xmc
from xmc.tools import dynamicImport
from xmc.tools import instantiateObject
from xmc.tools import convertObjectToFuture

# Import external libraries
import itertools as it
import numpy as np
import time 
import sys

class MonteCarloIndex():
    """
    This is a generic class for an index of an MXMC method. It handles sample
    generation and index-specific estimators.
    """

    def __init__(self,**keywordArgs):
        # Attributes
        self.indexValue = keywordArgs.get('indexValue')
        self.costEstimator = instantiateObject(keywordArgs.get('costEstimator'),**keywordArgs.get('costEstimatorInputDict'))
        self.qoiEstimator = []
        qoi_estimator_module = keywordArgs.get('qoiEstimator')
        qoi_estimator_module_args = keywordArgs.get('qoiEstimatorInputDict')
        for i in range(len(qoi_estimator_module)):
            self.qoiEstimator.append(instantiateObject(qoi_estimator_module[i],**qoi_estimator_module_args[i]))

        sampler_input_dict = keywordArgs.get('samplerInputDict')
        sampler_input_dict['solverWrapperIndices'] = self.solverIndices()
        self.sampler = instantiateObject(keywordArgs.get('sampler'),**sampler_input_dict)
        self.samples = keywordArgs.get('samples',None)
        self.areSamplesStored = keywordArgs.get('areSamplesStored',False)
        self.areSamplesRecycled = keywordArgs.get('areSampleRecycled',True)

    def sampleNumber(self):
        """
        Return the number of samples calculated from self.samples. 
        If no samples are stored, return number of samples stored
        in costEstimator.
        """
        if self.areSamplesStored:
            return len(self.samples)
        else:
            return self.costEstimator.sampleNumber()

    def indexwiseContribution(self, qoiEstimatorCoordinate, qoiEstimatorValueArguements):
        """
        For a given coordinate that uniquely identifies an entry of qoiEstimator, supply
        arguments to its value method and return the resultant estimation. 
        """
        return self.qoiEstimator[qoiEstimatorCoordinate].value(*qoiEstimatorValueArguements)

    def costMoment(self, costEstimatorValueArguments):
        """
        Returns the requested moment estimation of the cost. The input arguments are passed to the cost estimator's relevant method.
        """
        return self.costEstimator.value(*costEstimatorValueArguments)

    def newSample(self):
        """
        Request a new set of correlated samples. Calls its sampler as many times as necessary.
        """
        # Return 2^d QoI vectors that are generated from correlated underlying samples
        list_of_qoi = self.sampler.generate(self.indexValue)
        return list_of_qoi

    def indexSetDimension(self):
        """
        Return the current index set dimension computed from the length of self.indexValue
        """
        return len(self.indexValue)

    def solverIndices(self):
        """
        Compute list of indices of correlated solvers.
        """
        diff_tuple = []
        hypercube_vertices = 2**self.indexSetDimension()
        for _ in range(self.indexSetDimension()):
            diff_tuple.append([0,1])
        diff_list = list(it.product(*diff_tuple))
        for i in range(len(diff_list)):
            diff_list[i] = list(diff_list[i])

        main_index = [self.indexValue]*hypercube_vertices
        return [[main_index[i][j]-diff_list[i][j] for j in range(self.indexSetDimension())] for i in range(hypercube_vertices)]

    def _addSamples(self,newSamples):
        """
        Add newsamples to stored samples. This is an internal method with no check on whether sample storage is enabled; be careful when using it.
        """
        self.samples.append(newSamples)

    def update(self,newIndexAndSampleNumbers):
        """
        Update the Monte Carlo index to a new number of samples. First,
        decide the number of new samples to be generated, then generate
        each sample and the associated cost with newSample and pass them
        to the update methods of qoiEstimator and costEstimator.
        """
        # Compute number of new samples based on areSamplesRecycled
        if(self.areSamplesRecycled is True):
            number_new_samples = max(6, newIndexAndSampleNumbers[1] - self.sampleNumber())
        else:
            number_new_samples = max(6, newIndexAndSampleNumbers[1])
            ValueError("Computing samples from fresh not yet ready. Please set areSamplesRecycled to True")

        # Generate the required number correlated samples and add them into the 
        # corresponding power sums
        for _ in range(number_new_samples):
            list_of_qoi, total_time = self.newSample()

            # The size of this list should match with the first dimension of list_of_qoi
            assert (len(list_of_qoi)==len(self.qoiEstimator)),"qoiEstimator dimension not equal to list_of_qoi dimension"
            for i in range(len(list_of_qoi)):
                self.qoiEstimator[i].update(list_of_qoi[i])  # Update estimator-specific power sums

            self.costEstimator.update([total_time])
