# XMC imports
from xmc.tools import dynamicImport
import numpy as np

class MultiCriterion():
    """
    This class handles any number of criteria, the associated tolerances (with
    possible splitting), and the combination of these criteria into an interpred output.
    """

    def __init__(self,**keywordArgs):
        # Attributes
        self.criteria = keywordArgs.get('criteria')
        self.inputsForCriterion = keywordArgs.get('inputsForCriterion')
        self._interpreter = dynamicImport(keywordArgs.get('interpreter'))
        self.splitCriteria = keywordArgs.get('splitCriteria',None)
        self.toleranceToSplit = keywordArgs.get('toleranceToSplit',None)

        # Methods
        self._flagFunction =  dynamicImport(keywordArgs.get('flag', 'xmc.methodDefs_multiCriterion.flag.plainFlag'))

    def __getstate__(self):
        # Captures what is normally pickled
        state = self.__dict__.copy()
        # Replace the unserialisable PyCOMPSs-decorated entries with explicit label
        for attribute, value in state.items():
            if hasattr(value,'__module__') and 'pycompss' in value.__module__:
                state[attribute] = 'unassigned_task_decorator'
        # what we return here will be stored in the pickle
        return state


    def tolerances(self,criteriaReferences=None):
        """
        Returns the tolerances of the requested criteria.
        """
        if criteriaReferences is None:
            criteria_references = range(len(self.criteria))
        tolerance_values = [] 
        for i in criteriaReferences:
            tolerance_values.append(self.criteria[i].tolerance)
        return tolerance_values

    def splittingParameter(self):
        """
        Returns the splitting parameter currently applied
        """
        return self.criteria[self.splitCriteria[0]].tolerance/self.toleranceToSplit
                                    
    def setTolerance(self,criterionReferences,toleranceValues):
        """
        Accept a list of tolerances and set them inside each entry of self.criteria
        contained in criterionReferences
        """
        for i in range(len(criterionReferences)):
            self.criteria[criterionReferences[i]].tolerance = toleranceValues[i]

    def splitTolerance(self,splittingParameter):
        """
        Accept a splitting parameter and split the tolerance specified by toleranceToSplit. 
        Set this value inside the relevant self.criteria object
        """
        if self.toleranceToSplit is not None:
            split_tolerances = self.toleranceToSplit*np.array([splittingParameter, 1-splittingParameter])
            self.setTolerance(self.splitCriteria,split_tolerances)
        else:
            pass

    def flag(self,values):
        """
        Return the output of the criterion.
        It takes the expected values as input, evaluate each elementary criterion on them and combine their boolean outputs into a dictionary flag.
        This is currently a wrapper for the protected _flagFunction attribute.
        """
        return self._flagFunction(self,values)

    @staticmethod
    def flagStructure():
        """
        This function returns the typical output of a MultiCriterion's interpreter, 
        with default values. Such output is meant to be created by first calling 
        this function, then changing specific values according the interpreter. 
        The point is to always have the same data structure.
        """
        structure = {
            'stop': False, # stop algorithm
            'updateHierarchy': True,
            'updateTolerance': False, # for continuation-type algorithms
            'updateIndexSpace': False,
            'updateSampleNumberSpace': False,
        }
        return structure
