import importlib

# Import PyCOMPSs (if installed)
# from exaqute.ExaquteTaskPyCOMPSs import *   # to execute with runcompss

# Import empty definitions (if PyCOMPSs is not installed)
from scheduler.ExaquteTaskLocal import *      # to execute with python3

@ExaquteTask(returns=1)
def convertObjectToFuture(inputObject):
    """
    Convert the input object to a futures object.
    """
    return inputObject

def dynamicImport(fullName):
    """
    This function returns the requested attribute of a module; it manages the necessary imports.
    The input argument is the full name as a string, e.g. 'module.package.method'.
    Alternatively, the input argument may be a list of such full names. The output will then be a list of same size containing the requested attributes in the same order.
    This only works for modern Python (at least Python > 3).
    """
    if fullName is None: #safety
        return None
    elif fullName is list or fullName is tuple:
        requested_attributes = []
        for name in fullName:
            requested_attributes.append(dynamicImport(name))
        return requested_attributes
    else:
        split_name = fullName.split('.')
        module_name = '.'.join(split_name[0:-1])
        attribute_name = split_name[-1]
        module = importlib.import_module(module_name)
        requested_attribute = getattr(module,attribute_name)
        return requested_attribute
   
def instantiateObject(fullName, **kwargs):
    """
    Get the object handle described by the path in fullName and return an instance of it by 
    passing **kwargs to its constructor
    """
    object_type = dynamicImport(fullName)
    return object_type(**kwargs)
 
def doNothing(*args):
    """
    This is a function that does nothing. Useful to hold a place without taking any action.
    """
    pass

def returnInput(*args):
    """
    This is a function that returns exactly what is passed to it. Useful to hold a place in the dataflow without modifying the data.
    """
    return args

def getUnionAndMap(listOfLists):
    """
    Given a list of lists, computes union of all lists, as well as map from list of lists
    to union lists. 
    """
    union_list = []
    union_map = []
    for i in range(len(listOfLists)):
        union_map.append([None]*len(listOfLists[i]))
        for j in range(len(listOfLists[i])):
            found = False
            for k in range(len(union_list)):
                if listOfLists[i][j] == union_list[k]:
                    union_map[i][j] = k
                    found = True
                    break
            if found is False:
                union_list.append(listOfLists[i][j])
                union_map[i][j] = len(union_list)-1
    return union_list, union_map

def splitOneListIntoTwo(inputList):
    """
    Function that takes a list, each of whose entries is a list
    containing two entries, and assembles two separate lists, one
    for each of these entries
    """
    list1 = []
    list2 = []
    for entry in inputList:
        list1.append(entry[0])
        list2.append(entry[1])
    return list1, list2

def mergeTwoListsIntoOne(list1,list2):
    """
    Function that takes two lists and assembles them into one list
    such that each entry of the final list is a list containing the
    two corresponding entries of the two input lists
    """
    if (len(list1)!=len(list2)):
        raise ValueError('Input lists not of same length')

    output_list = [[list1[i],list2[i]] for i in range(len(list1))]
    return output_list

def strictlyPositiveBoundaryBooleans(indexSet):
    """
    Accepts a convex set of non-negative multi-indices and returns 
    a list of booleans of the same size that states whether the 
    corresponding index lies on the surface of the set where none
    of the indices have zero-values. 
    """
    max_indices = [max([indexSet[i][j] for i in range(len(indexSet))]) for j in range(len(indexSet[0])) ]
    is_index_in_bias = [None for _ in indexSet]
    for i in range(len(indexSet)):
        index = indexSet[i]
        is_index_max = [index[i]==max_indices[i] for i in range(len(index))]
        is_index_nonzero = [index[i]!=0 for i in range(len(index))]
        is_index_in_bias[i] = any(is_index_max) and all(is_index_nonzero)
    return is_index_in_bias

@ExaquteTask(returns=1)
def sum_Task(*args):
    """
    Return the sum of a list of futures objects as a futures object
    """
    return sum(packedList(args))

def packedList(obj):
    """
    Adapted from PyCOMPS module exaqute.ExaquteTask
    """
    index = [i for i, x in enumerate(obj) if x == "##"]
    if len(index) == 0:
        return obj
    new_vector = []
    new_vector.append(list(obj[0:index[0]]))
    for i in range(len(index) - 1):
        new_vector.append(list(obj[(index[i] + 1):index[i + 1]]))
    new_vector.append(list(obj[(index[-1] + 1):len(obj)]))
    print("Return:")
    print(new_vector)
    return new_vector


def unpackedList(obj):
    """
    Adapted from PyCOMPS module exaqute.ExaquteTask
    """
    if not any(isinstance(l,list) for l in obj):
        return obj
    new_vector = []
    new_vector.extend(obj[0])
    for i in range(1, len(obj)):
        new_vector.append("##")
        new_vector.extend(obj[1])
    return new_vector
