# XMC imports
from xmc.tools import dynamicImport
from xmc.tools import instantiateObject

# Import PyCOMPSs (if installed)
# from exaqute.ExaquteTaskPyCOMPSs import *   # to execute with runcompss

# Import empty definitions (if PyCOMPSs is not installed)
from scheduler.ExaquteTaskLocal import *      # to execute with python3


class XMCAlgorithm():
    """
    This top-level class handles the overall algorithm: initialisation as well
    as everything related to error, tolerance and iterations. It also possesses
    the necessary methods and attributes to create new types of algorithms. However,
    the export of results is to be handled outside. Although self-sufficient, this class may inherit from some metaclass.
    """

    def __init__(self, **keywordArgs):
        # Attributes
        self.monteCarloSampler = keywordArgs.get('monteCarloSampler')
        self.hierarchyOptimiser = keywordArgs.get('hierarchyOptimiser')
        self.stoppingCriterion = keywordArgs.get('stoppingCriterion')
        self.errorsForStoppingCriterion = keywordArgs.get('errorForStoppingCriterion',None)
        self.predictorsForHierarchy = keywordArgs.get('predictorsForHierarchy',None)
        self.estimatorsForHierarchy = keywordArgs.get('estimatorsForHierarchy',None)
        self.tolerancesForHierarchy = keywordArgs.get('tolerancesForHierarchy',None)
        self.costPredictorForHierarchy = keywordArgs.get('costPredictorForHierarchy',None)
        self.costEstimatorForHierarchy = keywordArgs.get('costEstimatorForHierarchy',None)
        self.iterationCounter = 0

        # Methods
        self.checkInitialisation = dynamicImport(keywordArgs.get('checkInitialisation','xmc.tools.doNothing'))
        self.updateTolerance = dynamicImport(keywordArgs.get('updateTolerance','xmc.tools.doNothing'))

    def indexEstimation(self, coordinate, valueMethodArgs):
        """
        Accept a coordinate that uniquely identifies an entry of qoiEstimator and the
        arguments that are required for the value method of this entry. Return a list
        that contains these index-wise estimations from each entry of 
        self.monteCarloSampler.indices
        """  
        return self.monteCarloSampler.indexEstimation(coordinate, valueMethodArgs)

    def indexCostEstimation(self, valueMethodArgs):
        """
        Accept arguments that are required for the value method of costEstimator. Return a list
        that contains these index-wise cost estimations from each entry of 
        self.monteCarloSampler.indices
        """
        return self.monteCarloSampler.indexCostEstimation(valueMethodArgs)

    def qoiPrediction(self):
        """
        Return the qoiPredictor object from self.monteCarloSampler
        """
        return self.monteCarloSampler.qoiPredictor

    def costPrediction(self):
        """
        Return the costPredictor object from self.monteCarloSampler
        """
        return self.monteCarloSampler.costPredictor

    def tolerances(self,*args):
        """
        Return the list of tolerances from self.StoppingCriterion
        """
        return self.stoppingCriterion.tolerances(*args)

    def hierarchy(self):
        """
        Returns current hierarchy of the MC estimator.
        """
        return self.monteCarloSampler.hierarchy()

    def splitTolerance(self, splittingParameter):
        """
        Method that interfaces with the MultiCriterion class to apply tolerance
        splitting
        """
        self.stoppingCriterion.splitTolerance(splittingParameter)

    def optimalHierarchy(self):
        """
        Method that interfaces with the HierarchyOptimiser class to compute
        the optimal hierarchy
        """
        old_hierarchy = self.hierarchy()
        input_dict = self.hierarchyOptimiser.inputDictionaryTemplate()
        if(self.iterationCounter == 0):
            newHierarchy = self.hierarchyOptimiser.defaultHierarchy
        else:
            # Build model_list and parameters_list
            model_list = []
            parameter_list = []
            qoi_predictors = self.qoiPrediction()
            for qoi_predictor_coordinate in self.predictorsForHierarchy:
                model_list.append(qoi_predictors[qoi_predictor_coordinate]._valueForParameters)
                parameter_list.append(qoi_predictors[qoi_predictor_coordinate].parameters)
            cost_predictor = self.costPrediction()

            # Build estimations_list
            estimation_list = []
            for qoi_estimator_coordinate_and_value_args in self.estimatorsForHierarchy:
                coordinate = qoi_estimator_coordinate_and_value_args[0]
                value_args = qoi_estimator_coordinate_and_value_args[1]
                estimation_list.append(self.indexEstimation(coordinate,value_args))
            if self.costEstimatorForHierarchy is not None:
                cost_estimations = self.indexCostEstimation(self.costEstimatorForHierarchy)
            else:
                cost_estimations = None

            input_dict['oldHierarchy'] = old_hierarchy
            input_dict['newSampleNumber'] = 25
            input_dict['cdfValue'] = 0.95
            input_dict['tolerances'] = self.tolerances(self.tolerancesForHierarchy)
            input_dict['models'] = model_list
            input_dict['parametersForModel'] = get_value_from_remote(parameter_list)
            input_dict['estimations'] = get_value_from_remote(estimation_list)
            if cost_predictor is None:
                input_dict['costModel'] = None
                input_dict['costParameters'] = None
            else:
                input_dict['costModel'] = cost_predictor._valueForParameters
                input_dict['costParameters'] = get_value_from_remote(cost_predictor.parameters)
            input_dict['costEstimations'] = get_value_from_remote(cost_estimations)

            newHierarchy = self.hierarchyOptimiser.optimalHierarchy(input_dict)
        return newHierarchy,input_dict.get('splittingParameter',None)

    def updateHierarchy(self,newHierarchy):
        """
        Method that interfaces with the monteCarloSample class to execute
        a given hierarchy
        """
        self.monteCarloSampler.update(newHierarchy)

    def estimation(self, assemblerCoordinates=None):
        """
        Method that calls the estimation method of monteCarloSampler
        """
        return self.monteCarloSampler.estimation(assemblerCoordinates)

    def errorEstimation(self, errorEstimatorCoordinates=None):
        """
        Method that calls the errorEstimation method of monteCarloSampler
        """
        return self.monteCarloSampler.errorEstimation(errorEstimatorCoordinates)

    def updateHierarchySpace(self,*args):
        """
        Method that interfaces with the HierarchyOptimiser class to compute
        the hierarchy space in which to search for the optimal hierarchy
        """
        self.hierarchyOptimiser.updateHierarchySpace(args)

    def stoppingCriterionFlag(self,currentCost=None):
        """
        Call stoppingCriterion.flag with the proper arguments and return its output
        (a.k.a flag).
        Input argument: currentCost is an indication of the cost the algorithm has entailed
        so far; we usually use the number of iterations.
        Output argument: criterion flag structure as define in the MultiCriterion class.
        """
        # Get errors required for stopping criterion
        errors = get_value_from_remote(self.errorEstimation(self.errorsForStoppingCriterion))

        # Set up dictionary required for stoppingCriterion.flag
        input_dictionary = {}
        for i in range(len(errors)):
            input_dictionary['error'+str(i)] = errors[i]
        input_dictionary['hierarchy'] = self.hierarchy()
        input_dictionary['algorithmCost'] = currentCost

        # Comput flag from dictionary and return
        flag = self.stoppingCriterion.flag(input_dictionary)
        return flag

    def runXMC(self):
        """
        Run an algorithm with generic structure, as described in the documentation.
        Other methods may be implemented in the future, for specific needs that do
        not fit even this generic algorithm.
        """

        self.checkInitialisation()

        # Iteration Loop will start here
        flag = self.stoppingCriterion.flagStructure()
        self.iterationCounter = 0
        while(not flag['stop']):
            newHierarchy,splittingParameter = self.optimalHierarchy()

            self.splitTolerance(splittingParameter)

            self.updateHierarchy(newHierarchy)

            # synchronization point needed to launch new tasks if convergence is false
            flag = self.stoppingCriterionFlag(self.iterationCounter)
            if flag['updateTolerance']:
                self.updateTolerance()
            if flag['updateIndexSpace']:
                self.updateHierarchySpace()

            self.iterationCounter += 1

            print("Iteration ",self.iterationCounter-1, "\tHierarchy - ", self.hierarchy(), "\tError - ",
                get_value_from_remote(self.errorEstimation(self.errorsForStoppingCriterion)))

        print("Estimation - ",get_value_from_remote(self.estimation()),
            "\nFinal Error - ",get_value_from_remote(self.errorEstimation(self.errorsForStoppingCriterion)),
            "\nLevelwise Mean Times - ",get_value_from_remote(self.indexCostEstimation([1,True,False])))
